/* ilbmtoppm.c - read an IFF ILBM file and produce a portable pixmap
**
** Copyright (C) 1989 by Jef Poskanzer.
**
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted, provided
** that the above copyright notice appear in all copies and that both that
** copyright notice and this permission notice appear in supporting
** documentation.  This software is provided "as is" without express or
** implied warranty.
**
** Modified by Mark Thompson on 10/4/90 to accomodate 24 bit IFF files
** as used by ASDG, NewTek, etc.
**
** Changed 20/Jun/93 by Ingo Wilken (Ingo.Wilken@informatik.uni-oldenburg.de)
**  - row-by-row operation
**  - better de-interleave algorithm,
**  - colormap files
**  - direct color
*/

#include "ppm.h"
#include "ilbm.h"


static void getfourchars ARGS(( FILE* f, char fourchars[4] ));
static unsigned char get_byte ARGS(( FILE* f ));
static long get_big_long ARGS((FILE *f));
static short get_big_short ARGS((FILE *f));
static void skip_chunk ARGS((FILE *f, long chunksize));
static pixel * read_colormap ARGS((FILE *f, long colors));
static BitMapHeader * read_bmhd ARGS((FILE *f));
static void readerr ARGS((FILE *f));
static void ham_to_ppm ARGS((FILE *ifp, BitMapHeader *bmhd, pixel *colormap, int colors));
static void deep_to_ppm ARGS((FILE *ifp, BitMapHeader *bmhd));
static void cmap_to_ppm ARGS((pixel *colormap, int colors));
static void std_to_ppm ARGS((FILE *ifp, BitMapHeader *bmhd, pixel *colormap, int colors, long viewportmodes));
static void direct_to_ppm ARGS((FILE *ifp, BitMapHeader *bmhd, DirectColor *dcol));
static pixel * ehb_to_cmap ARGS((pixel *colormap, int *colors));
static void read_ilbm_plane ARGS((FILE *ifp, int cols, int compression));
static void decode_row ARGS((FILE *ifp, rawtype *chunkyrow, int planes, BitMapHeader *bmhd));
static rawtype * alloc_rawrow ARGS((int cols));
static void * xmalloc ARGS((int bytes));
static void display_chunk ARGS((FILE *ifp, char *iffid, long chunksize));

static short verbose = 0;
static short do_mask = 0;
static unsigned char *ilbmrow;
static pixel *pixelrow;

void
main(argc, argv)
    int argc;
    char *argv[];
{
    FILE *ifp;
    pixel *colormap = 0;
    int argn, colors;
    char iffid[5];
    short body = 0;
    long formsize, bytesread, chunksize, viewportmodes = 0;
    char *usage = "[-verbose] [ilbmfile]";
    BitMapHeader *bmhd = NULL;
    DirectColor *dcol = NULL;

    ppm_init(&argc, argv);

    argn = 1;
    while( argn < argc && argv[argn][0] == '-' && argv[argn][1] != '\0' ) {
        if( pm_keymatch(argv[argn], "-verbose", 2) ) {
            verbose = 1;
        }
        else
        if( pm_keymatch(argv[argn], "-noverbose", 4) ) {
            verbose = 0;
        }
        else
            pm_usage(usage);
        ++argn;
    }

    if( argn < argc ) {
        ifp = pm_openr( argv[argn] );
        argn++;
    }
    else
        ifp = stdin;

    if( argn != argc )
        pm_usage(usage);

    /* Read in the ILBM file. */
    iffid[4] = '\0';
    getfourchars(ifp, iffid);
    if( strcmp(iffid, "FORM") != 0 )
        pm_error("input is not a FORM type IFF file");
    formsize = get_big_long(ifp);
    getfourchars(ifp, iffid);
    if( strcmp(iffid, "ILBM") != 0 )
        pm_error( "input is not an ILBM type FORM IFF file" );
    bytesread = 12;

    /* Main loop, parsing the IFF FORM. */
    while( bytesread < formsize ) {
        getfourchars(ifp, iffid);
        chunksize = get_big_long(ifp);
        bytesread += 8;

        if( body != 0 ) {
            pm_message("\"%s\" chunk found after BODY chunk - skipping", iffid);
            skip_chunk(ifp, chunksize);
        }
        else
        if( strcmp(iffid, "BMHD") == 0 )
            bmhd = read_bmhd(ifp);
        else
        if( strcmp(iffid, "CMAP") == 0 ) {
            colors = chunksize / 3;
            colormap = read_colormap(ifp, colors);
        }
        else
        if( strcmp(iffid, "CAMG") == 0 )
            viewportmodes = get_big_long(ifp);
        else
        if( strcmp(iffid, "DCOL") == 0 ) {
            dcol = xmalloc(sizeof(DirectColor));
            dcol->r = get_byte(ifp);
            dcol->g = get_byte(ifp);
            dcol->b = get_byte(ifp);
            (void) get_byte(ifp);
        }
        else
        if( strcmp(iffid, "BODY") == 0 ) {
            if( bmhd == NULL )
                pm_error("\"BODY\" chunk without \"BMHD\" chunk");

            ilbmrow = (unsigned char *)xmalloc(RowBytes(bmhd->w));
            pixelrow = ppm_allocrow(bmhd->w);

            if( viewportmodes & vmHAM )
                ham_to_ppm(ifp, bmhd, colormap, colors);
            else
            if( dcol != NULL )
                direct_to_ppm(ifp, bmhd, dcol);
            else
            if( bmhd->nPlanes == 24 )
                deep_to_ppm(ifp, bmhd);
            else
                std_to_ppm(ifp, bmhd, colormap, colors, viewportmodes);
            body = 1;
        }
        else
        if( strcmp(iffid, "GRAB") == 0 || strcmp(iffid, "DEST") == 0 ||
            strcmp(iffid, "SPRT") == 0 || strcmp(iffid, "CRNG") == 0 ||
            strcmp(iffid, "CCRT") == 0 || strcmp(iffid, "CLUT") == 0 ||
            strcmp(iffid, "DPPV") == 0 || strcmp(iffid, "DRNG") == 0 ||
            strcmp(iffid, "EPSF") == 0 ) {
            skip_chunk(ifp, chunksize);
        }
        else
        if( strcmp(iffid, "(c) ") == 0 || strcmp(iffid, "AUTH") == 0 ||
            strcmp(iffid, "NAME") == 0 || strcmp(iffid, "ANNO") == 0 ) {
            if( verbose )
                display_chunk(ifp, iffid, chunksize);
            else
                skip_chunk(ifp, chunksize);
        }
        else
        if( strcmp(iffid, "DPI ") == 0 ) {
            int x, y;

            x = get_big_short(ifp);
            y = get_big_short(ifp);
            if( verbose )
                pm_message("\"DPI \" chunk:  dpi_x = %d    dpi_y = %d", x, y);
        }
        else {
            pm_message("unknown chunk type \"%s\" - skipping", iffid);
            skip_chunk(ifp, chunksize);
        }

        bytesread += chunksize;
        if( odd(chunksize) ) {
                (void) get_byte(ifp);
                ++bytesread;
        }
    }
    pm_close(ifp);

    if( body == 0 ) {
        if( colormap )
            cmap_to_ppm(colormap, colors);
        else
            pm_error("no \"BODY\" or \"CMAP\" chunk found");
    }

    exit(0);
}


static void
readerr(f)
    FILE *f;
{
    if( ferror(f) )
        pm_error("read error");
    else
        pm_error("premature EOF");
}


static unsigned char
get_byte(ifp)
    FILE* ifp;
{
    int i;

    i = getc(ifp);
    if( i == EOF )
        readerr(ifp);

    return (unsigned char) i;
}

static void
getfourchars(ifp, fourchars)
    FILE* ifp;
    char fourchars[4];
{
    fourchars[0] = get_byte(ifp);
    fourchars[1] = get_byte(ifp);
    fourchars[2] = get_byte(ifp);
    fourchars[3] = get_byte(ifp);
}


static long
get_big_long(ifp)
    FILE *ifp;
{
    long l;

    if( pm_readbiglong(ifp, &l) == -1 )
        readerr(ifp);

    return l;
}

static short
get_big_short(ifp)
    FILE *ifp;
{
    short s;

    if( pm_readbigshort(ifp, &s) == -1 )
        readerr(ifp);

    return s;
}


static void
skip_chunk(ifp, chunksize)
    FILE *ifp;
    long chunksize;
{
    int i;

    for( i = 0; i < chunksize; i++ )
        (void) get_byte(ifp);
}


static pixel *
read_colormap(ifp, colors)
    FILE *ifp;
    long colors;
{
    pixel *colormap;
    int i, r, g, b;

    if( colors > 0 ) {
        colormap = ppm_allocrow(colors);
        for( i = 0; i < colors; i++ ) {
            r = get_byte(ifp);
            g = get_byte(ifp);
            b = get_byte(ifp);
            PPM_ASSIGN(colormap[i], r, g, b);
        }
    }
    else
        colormap = (pixel *)NULL;

    return colormap;
}


static BitMapHeader *
read_bmhd(ifp)
    FILE *ifp;
{
    BitMapHeader *bmhd;

    bmhd = (BitMapHeader *)xmalloc(sizeof(BitMapHeader));

    bmhd->w = get_big_short(ifp);           /* cols */
    bmhd->h = get_big_short(ifp);           /* rows */
    bmhd->x = get_big_short(ifp);
    bmhd->y = get_big_short(ifp);
    bmhd->nPlanes = get_byte(ifp);
    bmhd->masking = get_byte(ifp);
    bmhd->compression = get_byte(ifp);
    bmhd->pad1 = get_byte(ifp);             /* (ignored) */
    bmhd->transparentColor = get_big_short(ifp);
    bmhd->xAspect = get_byte(ifp);
    bmhd->yAspect = get_byte(ifp);
    bmhd->pageWidth = get_big_short(ifp);
    bmhd->pageHeight = get_big_short(ifp);

    /* fix aspect ratio */
    if( bmhd->xAspect == 0 ) {
        if( bmhd->yAspect == 0 ) {
            pm_message("warning - xAspect:yAspect = 0:0, using 1:1");
            bmhd->xAspect = bmhd->yAspect = 1;
        }
        else {
            pm_message("warning - xAspect = 0, setting to yAspect");
            bmhd->xAspect = bmhd->yAspect;
        }
    }
    else {
        if( bmhd->yAspect == 0 ) {
            pm_message("warning - yAspect = 0, setting to xAspect");
            bmhd->yAspect = bmhd->xAspect;
        }
    }
    if( bmhd->xAspect != bmhd->yAspect ) {
        pm_message("warning - non-square pixels; to fix do a 'pnmscale -%cscale %g'",
            bmhd->xAspect > bmhd->yAspect ? 'x' : 'y',
            bmhd->xAspect > bmhd->yAspect ? (float)(bmhd->xAspect)/bmhd->yAspect : (float)(bmhd->yAspect)/bmhd->xAspect);
    }

    return bmhd;
}


#define HAMCODE_CMAP      0     /* look up color in colormap */
#define HAMCODE_BLUE      1     /* new blue component */
#define HAMCODE_RED       2     /* new red component */
#define HAMCODE_GREEN     3     /* new green component */

static void
ham_to_ppm(ifp, bmhd, colormap, colors)
    FILE *ifp;
    BitMapHeader *bmhd;
    pixel *colormap;
    int colors;
{
    int cols, rows, hambits, hammask, col, row;
    pixval maxval;
    rawtype *rawrow;

    cols = bmhd->w;
    rows = bmhd->h;
    hambits = bmhd->nPlanes - 2;
    hammask = (1 << hambits) - 1;

    if( verbose )
        pm_message("input is a HAM%d file", bmhd->nPlanes);

    if( hambits > MAXPLANES )
        pm_error("too many planes (max %d)", MAXPLANES);
    if( hambits < 1 )
        pm_error("HAM requires 3 planes or more");

    maxval = pm_bitstomaxval(hambits);
    if( maxval > PPM_MAXMAXVAL )
        pm_error("nPlanes is too large - try reconfiguring with PGM_BIGGRAYS\n    or without PPM_PACKCOLORS" );

    /* scale colormap to new maxval */
    if( colormap && maxval != MAXCOLVAL ) {
        pixval r, g, b;
        int i;

        for( i = 0; i < colors; i++ ) {
            r = PPM_GETR(colormap[i]) * maxval / MAXCOLVAL;
            g = PPM_GETG(colormap[i]) * maxval / MAXCOLVAL;
            b = PPM_GETB(colormap[i]) * maxval / MAXCOLVAL;
            PPM_ASSIGN( colormap[i], r, g, b );
        }
    }

    rawrow = alloc_rawrow(cols);

    ppm_writeppminit(stdout, cols, rows, maxval, 0);
    for( row = 0; row < rows; row++ ) {
        pixval r, g, b;

        decode_row(ifp, rawrow, bmhd->nPlanes, bmhd);
        r = g = b = 0;
        for( col = 0; col < cols; col++ ) {
            switch((rawrow[col] >> hambits) & 0x03) {
                case HAMCODE_CMAP:
                    if( colormap && colors >= maxval )
                        pixelrow[col] = colormap[rawrow[col] & hammask];
                    else
                        PPM_ASSIGN(pixelrow[col], rawrow[col] & hammask,
                                   rawrow[col] & hammask, rawrow[col] & hammask);
                    r = PPM_GETR(pixelrow[col]);
                    g = PPM_GETG(pixelrow[col]);
                    b = PPM_GETB(pixelrow[col]);
                    break;
                case HAMCODE_BLUE:
                    b = rawrow[col] & hammask;
                    PPM_ASSIGN(pixelrow[col], r, g, b);
                    break;
                case HAMCODE_RED:
                    r = rawrow[col] & hammask;
                    PPM_ASSIGN(pixelrow[col], r, g, b);
                    break;
                case HAMCODE_GREEN:
                    g = rawrow[col] & hammask;
                    PPM_ASSIGN(pixelrow[col], r, g, b);
                    break;
                default:
                    pm_error("impossible HAM code");
            }
        }
        ppm_writeppmrow(stdout, pixelrow, cols, (pixval) maxval, 0);
    }
}


static void
deep_to_ppm(ifp, bmhd)
    FILE *ifp;
    BitMapHeader *bmhd;
{
    int cols, rows, col, row;
    rawtype *Rrow, *Grow, *Brow;

    cols = bmhd->w;
    rows = bmhd->h;

    if( verbose )
        pm_message("input is a deep (24bit) ILBM");

    Rrow = alloc_rawrow(cols);
    Grow = alloc_rawrow(cols);
    Brow = alloc_rawrow(cols);

    ppm_writeppminit(stdout, cols, rows, MAXCOLVAL, 0);
    for( row = 0; row < rows; row++ ) {
        decode_row(ifp, Rrow, 8, bmhd);
        decode_row(ifp, Grow, 8, bmhd);
        decode_row(ifp, Brow, 8, bmhd);
        for( col = 0; col < cols; col++ )
            PPM_ASSIGN(pixelrow[col], Rrow[col], Grow[col], Brow[col]);
        ppm_writeppmrow(stdout, pixelrow, cols, MAXCOLVAL, 0);
    }
    pm_close(stdout);
}


static void
direct_to_ppm(ifp, bmhd, dcol)
    FILE *ifp;
    BitMapHeader *bmhd;
    DirectColor *dcol;
{
    int cols, rows, col, row, redplanes, greenplanes, blueplanes;
    rawtype *Rrow, *Grow, *Brow;
    pixval maxval, redmaxval, greenmaxval, bluemaxval;
    int scale;

    cols = bmhd->w;
    rows = bmhd->h;

    redplanes = dcol->r; greenplanes = dcol->g; blueplanes = dcol->b;
    if( redplanes > MAXPLANES || blueplanes > MAXPLANES || greenplanes > MAXPLANES )
        pm_error("too many planes (max %d per color)", MAXPLANES);

    if( bmhd->nPlanes != (redplanes + greenplanes + blueplanes) )
        pm_error("BMHD/DCOL plane number mismatch");

    if( verbose )
        pm_message("input is a %d:%d:%d direct color ILBM",
                    redplanes, greenplanes, blueplanes);

    if( redplanes == blueplanes && redplanes == greenplanes ) {
        scale = 0;
        maxval = pm_bitstomaxval(redplanes);
    }
    else {
        scale = 1;
        redmaxval   = pm_bitstomaxval(redplanes);
        greenmaxval = pm_bitstomaxval(greenplanes);
        bluemaxval  = pm_bitstomaxval(blueplanes);

        maxval = max(redmaxval, max(greenmaxval, bluemaxval));
        pm_message("rescaling colors to %d bits", pm_maxvaltobits(maxval));
    }

    if( maxval > PPM_MAXMAXVAL )
        pm_error("too many planes - try reconfiguring with PGM_BIGGRAYS\n    or without PPM_PACKCOLORS" );

    Rrow = alloc_rawrow(cols);
    Grow = alloc_rawrow(cols);
    Brow = alloc_rawrow(cols);

    ppm_writeppminit(stdout, cols, rows, maxval, 0);
    for( row = 0; row < rows; row++ ) {
        decode_row(ifp, Rrow, dcol->r, bmhd);
        decode_row(ifp, Grow, dcol->g, bmhd);
        decode_row(ifp, Brow, dcol->b, bmhd);

        if( scale ) {
            for( col = 0; col < cols; col++ ) {
                PPM_ASSIGN(pixelrow[col],
                            Rrow[col] * maxval / redmaxval,
                            Grow[col] * maxval / greenmaxval,
                            Brow[col] * maxval / bluemaxval);
            }
        }
        else {
            for( col = 0; col < cols; col++ )
                PPM_ASSIGN(pixelrow[col], Rrow[col], Grow[col], Brow[col]);
        }
        ppm_writeppmrow(stdout, pixelrow, cols, maxval, 0);
    }
    pm_close(stdout);
}


static void
cmap_to_ppm(colormap, colors)
    pixel *colormap;
    int colors;
{
    pm_message("input is a colormap file");

    ppm_writeppminit(stdout, colors, 1, MAXCOLVAL, 0);
    ppm_writeppmrow(stdout, colormap, colors, MAXCOLVAL, 0);
    pm_close(stdout);
}


static void
std_to_ppm(ifp, bmhd, colormap, colors, viewportmodes)
    FILE *ifp;
    BitMapHeader *bmhd;
    pixel *colormap;
    int colors;
    long viewportmodes;
{
    rawtype *rawrow;
    pixval maxval;
    int row, rows, col, cols;

    cols = bmhd->w;
    rows = bmhd->h;

    if( verbose )
        pm_message("input is a %d-plane %s ILBM", bmhd->nPlanes,
                    viewportmodes & vmEXTRA_HALFBRITE ? "EHB" : "standard");

    if( bmhd->nPlanes > MAXPLANES )
        pm_error("too many planes (max %d)", MAXPLANES);

    if( colormap )
        maxval = MAXCOLVAL;
    else {
        maxval = pm_bitstomaxval(bmhd->nPlanes);
        pm_message("no colormap - interpreting values as grayscale");
    }
    if( maxval > PPM_MAXMAXVAL )
        pm_error("nPlanes is too large - try reconfiguring with PGM_BIGGRAYS\n    or without PPM_PACKCOLORS" );

    rawrow = alloc_rawrow(cols);

    if( viewportmodes & vmEXTRA_HALFBRITE )
        colormap = ehb_to_cmap(colormap, &colors);

    ppm_writeppminit( stdout, cols, rows, (pixval) maxval, 0 );
    for( row = 0; row < rows; row++ ) {
        decode_row(ifp, rawrow, bmhd->nPlanes, bmhd);
        for( col = 0; col < cols; col++ ) {
            if( colormap )
                pixelrow[col] = colormap[rawrow[col]];
            else
                PPM_ASSIGN(pixelrow[col], rawrow[col], rawrow[col], rawrow[col]);
        }
        ppm_writeppmrow(stdout, pixelrow, cols, maxval, 0);
    }
}


static pixel *
ehb_to_cmap(colormap, colors)
    pixel *colormap;
    int *colors;
{
    pixel *tempcolormap;
    int i, col;

    col = *colors;
    tempcolormap = ppm_allocrow(col * 2);
    for( i = 0; i < col; i++ ) {
        tempcolormap[i] = colormap[i];
        PPM_ASSIGN(tempcolormap[col + i], PPM_GETR(colormap[i]) / 2,
                    PPM_GETG(colormap[i]) / 2, PPM_GETB(colormap[i]) / 2 );
    }
    ppm_freerow(colormap);
    *colors *= 2;

    return tempcolormap;
}


static void
read_ilbm_plane(ifp, cols, compression)
    FILE *ifp;
    int cols, compression;
{
    unsigned char *ubp;
    int bytes, j, byte;

    bytes = RowBytes(cols);

    switch(compression) {
        case cmpNone:
            j = fread(ilbmrow, 1, bytes, ifp);
            if( j != bytes )
                readerr(ifp);
            break;
        case cmpByteRun1:
            ubp = ilbmrow;
            do {
                byte = (int)get_byte(ifp);
                if( byte <= 127 ) {
                    j = byte;
                    bytes -= (j+1);
                    if( bytes < 0 )
                        pm_error("error doing ByteRun1 decompression");
                    for( ; j >= 0; j-- )
                        *ubp++ = get_byte(ifp);
                }
                else
                if ( byte != 128 ) {
                    j = 256 - byte;
                    bytes -= (j+1);
                    if( bytes < 0 )
                        pm_error("error doing ByteRun1 decompression");
                    byte = (int)get_byte(ifp);
                    for( ; j >= 0; j-- )
                        *ubp++ = (unsigned char)byte;
                }
                /* 128 is a NOP */
            }
            while( bytes > 0 );
            break;
        default:
                pm_error("unknown compression type");
    }
}


const unsigned char bit_mask[] = {1, 2, 4, 8, 16, 32, 64, 128};

static void
decode_row(ifp, chunkyrow, nPlanes, bmhd)
    FILE *ifp;
    rawtype *chunkyrow;
    int nPlanes;
    BitMapHeader *bmhd;
{
    int plane, col, cols;
    unsigned char *ilp;
    rawtype *chp;

    cols = bmhd->w;
    for( plane = 0; plane < nPlanes; plane++ ) {
        int mask, bit;

        mask = 1 << plane;
        read_ilbm_plane(ifp, cols, bmhd->compression);

        ilp = ilbmrow;
        chp = chunkyrow;

        bit = 7;
        for( col = 0; col < cols; col++, bit--, chp++ ) {
            if( bit < 0 ) {
                bit = 7;
                ilp++;
            }
            if( *ilp & bit_mask[bit] )
                *chp |= mask;
            else
                *chp &= ~mask;
        }
    }
    /* skip mask plane */
    if( bmhd->masking == mskHasMask )
        read_ilbm_plane(ifp, cols, bmhd->compression);
}


static rawtype *
alloc_rawrow(cols)
    int cols;
{
    rawtype *r;
    int i;

    r = (rawtype *)xmalloc(cols * sizeof(rawtype));

    for( i = 0; i < cols; i++ )
        r[i] = 0;

    return r;
}


static void *
xmalloc(bytes)
    int bytes;
{
    void *mem;

    mem = malloc(bytes);
    if( mem == NULL )
        pm_error("out of memory allocating %d bytes", bytes);
    return mem;
}


static void
display_chunk(ifp, iffid, chunksize)
    FILE *ifp;
    char *iffid;
    long chunksize;
{
    int byte;

    pm_message("contents of \"%s\" chunk:", iffid);

    while( chunksize-- ) {
        byte = get_byte(ifp);
        if( fputc(byte, stderr) == EOF )
            pm_error("write error");
    }
    if( fputc('\n', stderr) == EOF )
        pm_error("write error");
}

